(* use grammars.sml *)

Example 1:
---------
structure R = 
    ppGRAMMARS(fun varName i = if i=6 then "x" 
		               else "y("^ Int.toString i ^")");

let
    open R
in
    printExp (Star (Plus[Atom "a",
			 Star (Star(Times[Var 8,Atom "B"])), 
			 Star(Var 6),
			 Atom "e"]));
    printEqns [(1,Star (Plus [Atom "ab",Var 0])),
	       (0,Times[Atom "ba",Plus [Var 0,Var 1]]),
	       (2,Times[Atom "aba",Plus [Var 10]])]
end;


Example 2: (checking containment of the empty word)
----------
val falseEnv = (fn (i:int) => false);
val trueEnv = (fn (i:int) => true);

structure Grm = ppREGEXP (StandardNames);
open Grm;

val Eqns = [(0,Times[Var 3,Var 2, Var 4]),
	    (1,Plus[Times[Atom "b",Var 2],Atom "c"]),
	    (3,Plus[Times[Atom "a",Var 3,Atom "b"],Times[]])];

printEqns Eqns;
 Equation system: 
   X0 = X3;X2;X4
   X1 = (b;X2 + c)
   X3 = (a;X3;b + 1)

Grammars.geUnits falseEnv Eqns;
val it = [(0,false),(1,false),(3,true)] : (int * bool) list
Grammars.geUnits (fn i => if 0 < i andalso i < 4 then true else false) Eqns;
val it = [(0,false),(1,false),(3,true)] : (int * bool) list
Grammars.geUnits (fn i => if i=2 orelse i=4 then true else false) Eqns;
val it = [(0,true),(1,false),(3,true)] : (int * bool) list


Example 3: (E-normal form of a grammar: set all empty components to 0)
----------
val Eqns = [(0,Times[Var 3,Var 2, Var 4]),
	    (1,Plus[Times[Atom "b",Var 2]]),
	    (3,Plus[Times[Atom "a",Var 0,Var 3,Var 4,Atom "b"],Times[]])];
printEqns Eqns;
 Equation system: 
   X0 = X3;X2;X4
   X1 = b;X2
   X3 = (a;X0;X3;X4;b + 1)

printEqns (Grammars.nfEmpty (fn i => i=4) Eqns);
 Equation system: 
   X0 = 0
   X1 = b;X2
   X3 = 1
printEqns (Grammars.nfEmpty (fn i => i=2) Eqns);
 Equation system: 
   X0 = 0
   X1 = 0
   X3 = 1


Example 4: Subtraction of 1 in all components of a regular system
---------
(* Aufgabe 6.1, S.116  Automatentheorie und Formale Sprachen *)

structure R = ppGRAMMARS(
		fun varName 0 = "A" 
		  | varName 1 = "B" 
		  | varName 2 = "C"
                  | varName 3 = "D" 
		  | varName i = if 2<i andalso i<12 
				    then "Y("^(Int.toString (Int.div(i,3)))^","
                                          ^(Int.toString(1+Int.rem(i,3)))^")"
				else "V"^(Int.toString i)
              );
open R;

val eqs : string re Grammars.eqns 
        = [(0,Times [Var 1, Var 2, Var 3]),
	   (1,Plus [Times [Atom "b", Var 2], Atom "c"]),
           (2,Plus [Times [Var 2, Atom "c"], Times[Var 3, Var 3]]),
           (3,Plus [Times [Atom "a", Var 3, Atom "b"], Unit])
          ];
printEqns eqs;
 Equation system: 
   A = B;C;D
   B = (b;C + c)
   C = (C;c + D;D)
   D = (a;D;b + 1)
val bools = Grammars.geUnits (fn i => false) eqs;
(* val bools = [(0,false),(1,false),(2,true),(3,true)]    ok *)

val eqsmin = 
    map (fn (i,r) => (i,RegExp.dnf r)) (Grammars.subtractUnits (fn i => false) eqs);
printEqns eqsmin;
 Equation system: 
   A = (B + B;D + B;C;D + B;C)
   B = (c + b + b;C)
   C = (c + C;c + D + D;D)
   D = (a;b + a;D;b)
val boolsmin = Grammars.geUnits (fn i => false) eqsmin;
val boolsmin = [(0,false),(1,false),(2,false),(3,false)] : (int * bool) list

(* Variante mit c* : *)
val eqs = [(0,Times [Var 1, Var 2, Var 3]),
	   (1,Plus [Times [Atom "b", Var 2], Atom "c"]),
	   (2,Plus [Times [Var 3, Var 3, Star(Atom "c")]]),
	   (3,Plus [Times [Atom "a", Var 3, Atom "b"], Unit])
         ];
printEqns eqs;
 Equation system: 
   A = B;C;D
   B = (b;C + c)
   C = D;D;(c)*
   D = (a;D;b + 1)
val eqsmin = map (fn (i,r) => 
             (i,RegExp.dnf (Grammars.subtractUnit (fn i => false) eqs r))) eqs;
printEqns eqsmin;
 Equation system: 
   A = (B + B;D + B;C;D + B;C)
   B = (c + b + b;C)
   C = (c;(c)* + D;c;(c)* + D + D;D;c;(c)* + D;D)
   D = (a;b + a;D;b)

(* Was soll mit nicht-Rekursionsvariablen geschehen? Oben wird deren Bedeutung
   von X zu (X-1) geaendert, aber was bedeutet das fuer die definierten Sprachen?
   Sicherer waere, anzunehmen, dass env(X) = 0 besagt, dass 1 nicht in X liegt.

val eqs = [(0,Times [Var 4]),
           (2,Plus [Times [Var 2, Atom "c"], Times[Var 3, Var 3]]),
           (3,Plus [Times [Atom "a", Var 3, Atom "b"], Unit])
         ];
val bools = map (Grammars.geUnit (fn i => not (i=4)) eqs) (map #2 eqs);
val bools = [false,true,true] : bool list

val eqsmin = map (fn (i,r) => 
                 (i,RegExp.dnf (Grammars.subtractUnit (fn i => false) eqs r))) eqs;
printEqns eqsmin;
  [(0,Var 4),
   (2,Plus [Atom "c",Times [Var 2,Atom "c"],Var 3,Times [Var 3,Var 3]]),
   (3,Plus [Times [Atom "a",Atom "b"],Times [Atom "a",Var 3,Atom "b"]])]

val boolsmin = map (Grammars.geUnit (fn i => false) eqsmin) (map #2 eqsmin);
val boolsmin = [false,false,false] : bool list

val eqs = [(0,Plus[Var 4,Times[]]),(3,Plus [Times [Atom "a", Var 3, Atom "b"], Unit])];
val bools = map (Grammars.geUnit (fn i => not(i=4)) eqs) (map #2 eqs);
val bools = [true,true] : bool list
val eqsmin = map (fn (i,r) => 
                  (i,RegExp.dnf (Grammars.subtractUnit (fn i => false) eqs r))) eqs;
val boolsmin = map (Grammars.geUnit (fn i => false) eqsmin) (map #2 eqsmin);
val boolsmin = [false,false] : bool list


Example 5: Elimination of left recursion in context free grammars
---------
(* Aufgabe 6.2 *)

val (G,R,T,elrG,GNF) =
let 
    val (NP,AP,N,A,AV) = (Var 0,Var 1,Var 2,Var 3,Var 4)
    val (die,und,sehr,Studenten,Aufgaben,leichte,schwierige,Probleme,loesenden)
	= (Atom "die",Atom "und",Atom "sehr",Atom "Studenten",
	   Atom "Aufgaben", Atom "leichte",
	   Atom "schwierige", Atom "Probleme", Atom "loesenden")
    val G = [(0,Plus[Times[die,AP,N],Times[AP,N],Times[NP,und,NP]]),
	     (1,Plus[Times[sehr,A],Times[NP,AV],A]),
	     (2,Plus[Studenten,Aufgaben,Probleme]),
	     (3,Plus[leichte,schwierige]),
	     (4,loesenden)]
    val (R,T) = RegEqns.isolateLeftVarFactors G 
    val elrG = RegEqns.leftToRightRecursion G 
    val GNF = RegEqns.greibachNF G 
in
    (G,R,T,elrG,GNF) 
end;

structure ppRegEqns = 
    ppGRAMMARS(val varName = 
		 (fn 0 => "NP" | 1 => "AP" | 2 => "N" | 3 => "A" | 4 => "AV"
  	           | n => if 4<n then 
		          "Y("^Int.toString(Int.div(n,5))^","
			      ^Int.toString(Int.rem(n,5)+1)^")"
			  else "V("^ Int.toString n ^")")
            );
open ppRegEqns;

printEqns G;
 Equation system: 
   NP = (die;AP;N + AP;N + NP;und;NP)
   AP = (sehr;A + NP;AV + A)
   N = (Studenten + Aufgaben + Probleme)
   A = (leichte + schwierige)
   AV = loesenden

printEqns elrG;

 Equation system: 
   NP = (die;AP;N + loesenden;Y(5,1) + leichte;Y(4,1) + schwierige;Y(4,1) 
         + Studenten;Y(3,1) + Aufgaben;Y(3,1) + Probleme;Y(3,1) + die;AP;N;Y(1,1) 
         + sehr;A;Y(2,1))
   AP = (sehr;A + loesenden;Y(5,2) + leichte;Y(4,2) + schwierige;Y(4,2) 
         + Studenten;Y(3,2) + Aufgaben;Y(3,2) + Probleme;Y(3,2) + die;AP;N;Y(1,2) 
         + sehr;A;Y(2,2))
   N = (loesenden;Y(5,3) + leichte;Y(4,3) + schwierige;Y(4,3) + Studenten;Y(3,3) 
         + Aufgaben;Y(3,3) + Probleme;Y(3,3) + die;AP;N;Y(1,3) + sehr;A;Y(2,3) 
         + Studenten + Aufgaben + Probleme)
   A = (loesenden;Y(5,4) + leichte;Y(4,4) + schwierige;Y(4,4) + Studenten;Y(3,4) 
         + Aufgaben;Y(3,4) + Probleme;Y(3,4) + die;AP;N;Y(1,4) + sehr;A;Y(2,4) 
         + leichte + schwierige)
   AV = (loesenden + loesenden;Y(5,5) + leichte;Y(4,5) + schwierige;Y(4,5) 
         + Studenten;Y(3,5) + Aufgaben;Y(3,5) + Probleme;Y(3,5) + die;AP;N;Y(1,5) 
         + sehr;A;Y(2,5))
   Y(1,1) = (und;NP + und;NP;Y(1,1) + AV;Y(2,1))
   Y(1,2) = (AV + und;NP;Y(1,2) + AV;Y(2,2))
   Y(1,3) = (und;NP;Y(1,3) + AV;Y(2,3))
   Y(1,4) = (und;NP;Y(1,4) + AV;Y(2,4))
   Y(1,5) = (und;NP;Y(1,5) + AV;Y(2,5))
   Y(2,1) = (N;Y(1,1) + N)
   Y(2,2) = N;Y(1,2)
   Y(2,3) = N;Y(1,3)
   Y(2,4) = N;Y(1,4)
   Y(2,5) = N;Y(1,5)
   Y(3,1) = 0
   Y(3,2) = 0
   Y(3,3) = 0
   Y(3,4) = 0
   Y(3,5) = 0
   Y(4,1) = Y(2,1)
   Y(4,2) = (Y(2,2) + 1)
   Y(4,3) = Y(2,3)
   Y(4,4) = Y(2,4)
   Y(4,5) = Y(2,5)
   Y(5,1) = 0
   Y(5,2) = 0
   Y(5,3) = 0
   Y(5,4) = 0
   Y(5,5) = 0

Wenn man Summanden mit Faktor 0 loescht, bekommt man folgende Greibach-Normalform:

val eGNF = (Grammars.nfEmpty (fn i => false)) GNF;
printEqns eGNF;

 Equation system: 
   NP = (die;AP;N + (leichte + schwierige);Y(4,1) + die;AP;N;Y(1,1) + sehr;A;Y(2,1))
   AP = (sehr;A + (leichte + schwierige);Y(4,2) + die;AP;N;Y(1,2) + sehr;A;Y(2,2))
   N  = ((leichte + schwierige);Y(4,3) + die;AP;N;Y(1,3) + sehr;A;Y(2,3) + Studenten 
         + Aufgaben + Probleme)
   A  = ((leichte + schwierige);Y(4,4) + die;AP;N;Y(1,4) + sehr;A;Y(2,4) + leichte 
         + schwierige)
   AV = (loesenden + (leichte + schwierige);Y(4,5) + die;AP;N;Y(1,5) + sehr;A;Y(2,5))
   Y(1,1) = (und;die;AP;N + und;leichte;Y(4,1) + und;schwierige;Y(4,1) 
          + und;die;AP;N;Y(1,1) + und;sehr;A;Y(2,1) + und;leichte;Y(4,1);Y(1,1) 
          + und;schwierige;Y(4,1);Y(1,1) + und;die;AP;N;Y(1,1);Y(1,1) 
          + und;sehr;A;Y(2,1);Y(1,1) + loesenden;Y(2,1) + leichte;Y(4,5);Y(2,1)
          + schwierige;Y(4,5);Y(2,1) + die;AP;N;Y(1,5);Y(2,1) + sehr;A;Y(2,5);Y(2,1))
   Y(1,2) = (loesenden + leichte;Y(4,5) + schwierige;Y(4,5) + die;AP;N;Y(1,5) 
          + sehr;A;Y(2,5) + und;die;AP;N;Y(1,2) + und;leichte;Y(4,1);Y(1,2) 
          + und;schwierige;Y(4,1);Y(1,2) + und;die;AP;N;Y(1,1);Y(1,2) 
          + und;sehr;A;Y(2,1);Y(1,2) + loesenden;Y(2,2) + leichte;Y(4,5);Y(2,2) 
          + schwierige;Y(4,5);Y(2,2) + die;AP;N;Y(1,5);Y(2,2) + sehr;A;Y(2,5);Y(2,2))
   Y(1,3) = (und;die;AP;N;Y(1,3) + und;leichte;Y(4,1);Y(1,3) 
          + und;schwierige;Y(4,1);Y(1,3) + und;die;AP;N;Y(1,1);Y(1,3) 
          + und;sehr;A;Y(2,1);Y(1,3) + loesenden;Y(2,3) + leichte;Y(4,5);Y(2,3) 
          + schwierige;Y(4,5);Y(2,3) + die;AP;N;Y(1,5);Y(2,3) + sehr;A;Y(2,5);Y(2,3))
   Y(1,4) = (und;die;AP;N;Y(1,4) + und;leichte;Y(4,1);Y(1,4) 
          + und;schwierige;Y(4,1);Y(1,4) + und;die;AP;N;Y(1,1);Y(1,4) 
          + und;sehr;A;Y(2,1);Y(1,4) + loesenden;Y(2,4) + leichte;Y(4,5);Y(2,4) 
          + schwierige;Y(4,5);Y(2,4) + die;AP;N;Y(1,5);Y(2,4) + sehr;A;Y(2,5);Y(2,4))
   Y(1,5) = (und;die;AP;N;Y(1,5) + und;leichte;Y(4,1);Y(1,5) 
          + und;schwierige;Y(4,1);Y(1,5) + und;die;AP;N;Y(1,1);Y(1,5) 
          + und;sehr;A;Y(2,1);Y(1,5) + loesenden;Y(2,5) + leichte;Y(4,5);Y(2,5) 
          + schwierige;Y(4,5);Y(2,5) + die;AP;N;Y(1,5);Y(2,5) + sehr;A;Y(2,5);Y(2,5))
   Y(2,1) = (leichte;Y(4,3);Y(1,1) + schwierige;Y(4,3);Y(1,1) 
          + die;AP;N;Y(1,3);Y(1,1) + sehr;A;Y(2,3);Y(1,1) + Studenten;Y(1,1) 
          + Aufgaben;Y(1,1) + Probleme;Y(1,1) + leichte;Y(4,3) + schwierige;Y(4,3) 
          + die;AP;N;Y(1,3) + sehr;A;Y(2,3) + Studenten + Aufgaben + Probleme)
   Y(2,2) = (leichte;Y(4,3);Y(1,2) + schwierige;Y(4,3);Y(1,2) 
          + die;AP;N;Y(1,3);Y(1,2) + sehr;A;Y(2,3);Y(1,2) + Studenten;Y(1,2) 
          + Aufgaben;Y(1,2) + Probleme;Y(1,2))
   Y(2,3) = (leichte;Y(4,3);Y(1,3) + schwierige;Y(4,3);Y(1,3) 
          + die;AP;N;Y(1,3);Y(1,3) + sehr;A;Y(2,3);Y(1,3) + Studenten;Y(1,3) 
          + Aufgaben;Y(1,3) + Probleme;Y(1,3))
   Y(2,4) = (leichte;Y(4,3);Y(1,4) + schwierige;Y(4,3);Y(1,4) 
          + die;AP;N;Y(1,3);Y(1,4) + sehr;A;Y(2,3);Y(1,4) + Studenten;Y(1,4) 
          + Aufgaben;Y(1,4) + Probleme;Y(1,4))
   Y(2,5) = (leichte;Y(4,3);Y(1,5) + schwierige;Y(4,3);Y(1,5) 
          + die;AP;N;Y(1,3);Y(1,5) + sehr;A;Y(2,3);Y(1,5) + Studenten;Y(1,5) 
          + Aufgaben;Y(1,5) + Probleme;Y(1,5))
   Y(3,1) = 0
   Y(3,2) = 0
   Y(3,3) = 0
   Y(3,4) = 0
   Y(3,5) = 0
   Y(4,1) = Y(2,1)
   Y(4,2) = (Y(2,2) + 1)
   Y(4,3) = Y(2,3)
   Y(4,4) = Y(2,4)
   Y(4,5) = Y(2,5)
   Y(5,1) = 0
   Y(5,2) = 0
   Y(5,3) = 0
   Y(5,4) = 0
   Y(5,5) = 0

(*
val prVar = (fn 0 => "NP" | 1 => "AP" | 2 => "N" | 3 => "A" | 4 => "AV"
              | n => if 4<n then "Y("^Int.toString(Int.div(n,5))^","
                                     ^Int.toString(Int.rem(n,5)+1)^")"
		     else "V("^ Int.toString n ^")")
and ppre = RegExp.pp prVar and ppreqns = RegEqns.pp prVar;

Compiler.Control.Print.printDepth := 40;
Compiler.Control.Print.printLength := 40;
Compiler.Control.Print.stringDepth := 700;

RegEqns.ppRegMat prVar R;
  #[#["und;NP","AV","0","0","0"],
    #["N","0","0","0","0"],
    #["0","0","0","0","0"],
    #["0","1","0","0","0"],
    #["0","0","0","0","0"]] : string vector vector

RegEqns.ppRegMat prVar T;
  #[#["die;AP;N","sehr;A","Studenten + Aufgaben + Probleme",
     "leichte + schwierige","loesenden"]] : string vector vector
*)

Teil (ii): Teilgrammatik:
------------------------
val (G,R,T,elrG,GNF) = 
let 
    val (NP,AP,N,A,AV) = (Var 0,Var 1,Atom "N",Atom "A",Atom "AV")
    val (die,und,sehr,Studenten,Aufgaben,leichte,schwierige,Probleme,loesenden)
	= (Atom "die",Atom "und",Atom "sehr",Atom "Studenten",
	   Atom "Aufgaben", Atom "leichte",
	   Atom "schwierige", Atom "Probleme", Atom "loesenden")
    val G = [(0,Plus[Times[die,AP,N],Times[AP,N],Times[NP,und,NP]]),
	     (1,Plus[Times[sehr,A],Times[NP,AV],A])]
    val (R,T) = RegEqns.isolateLeftVarFactors G 
    val elrG = RegEqns.leftToRightRecursion G 
    val GNF = RegEqns.greibachNF G 
in
    (G,R,T,elrG,GNF) 
end;

(*
val Y = RegMat.tabulate (2,2) 
        (fn (0,0) => Var 2 | (0,1) => (Var 3) | (1,0) => Var 4
          | (1,1) => Var 5 | _ => raise Undefined) : string RegMat.t
val TY = RegMat.prod(T,Y);
val TYT = RegMat.sum(RegMat.prod(T,Y),T);

val prVar = (fn 0 => "NP" | 1 => "AP" 
              | n => if n<6 then "Y("^Int.toString(Int.div(n,2))^","
                                     ^Int.toString(Int.rem(n,2)+1)^")"
		     else "V("^ Int.toString n ^")")

RegEqns.ppRegMat prVar TY;
RegEqns.ppRegMat prVar TYT;

RegEqns.ppRegMat prVar R;
val it = #[#["und;NP","AV"],
	   #["N","0"]] : string vector vector
RegEqns.ppRegMat prVar T;
val it = #[#["die;AP;N","A + sehr;A"]] : string vector vector
bzw:
val it = #[#["die;AP;N","sehr;A + A"]] : string vector vector
*)

structure ppRegEqns = 
    ppGRAMMARS(val varName = 
	    (fn 0 => "NP" | 1 => "AP" 
              | n => if n<6 then "Y("^Int.toString(Int.div(n,2))^","
                                     ^Int.toString(Int.rem(n,2)+1)^")"
		     else "V("^ Int.toString n ^")"));
open ppRegEqns;

printEqns G;
 Equation system: 
   NP = (die;AP;N + AP;N + NP;und;NP)
   AP = (sehr;A + NP;AV + A)

printEqns elrG;
 Equation system: 
   NP = (die;AP;N + die;AP;N;Y(1,1) + sehr;A;Y(2,1) + A;Y(2,1))
   AP = (die;AP;N;Y(1,2) + sehr;A;Y(2,2) + A;Y(2,2) + sehr;A + A)
   Y(1,1) = (und;NP + und;NP;Y(1,1) + AV;Y(2,1))
   Y(1,2) = (AV + und;NP;Y(1,2) + AV;Y(2,2))
   Y(2,1) = (N;Y(1,1) + N)
   Y(2,2) = N;Y(1,2)

printEqns GNF;
 Equation system: 
   NP = (die;AP;N + die;AP;N;Y(1,1) + (sehr;A + A);Y(2,1))
   AP = (die;AP;N;Y(1,2) + (sehr;A + A);Y(2,2) + sehr;A + A)
   Y(1,1) = (und;die;AP;N + und;die;AP;N;Y(1,1) + und;sehr;A;Y(2,1) + und;A;Y(2,1) 
             + AV;Y(2,1) + und;die;AP;N;Y(1,1);Y(1,1) + und;sehr;A;Y(2,1);Y(1,1) 
             + und;A;Y(2,1);Y(1,1))
   Y(1,2) = (AV + AV;Y(2,2) + und;die;AP;N;Y(1,2) + und;die;AP;N;Y(1,1);Y(1,2) 
             + und;sehr;A;Y(2,1);Y(1,2) + und;A;Y(2,1);Y(1,2))
   Y(2,1) = (N;Y(1,1) + N)
   Y(2,2) = N;Y(1,2)

(* Man sollte die alten Variablen nur an fuehrenden Vorkommen in Summanden 
   der neuen Gleichungen ersetzen. *)


(* --------------- Precomputations for LR(0) Grammars ---------------- *)

(*      Left contexts of a variable in all categories of a grammar     *)

Example 1: (Arbib, Moll, Kfoury p.136)

let
    open Grm
    open Grammars
    val eqs = [(0,Plus [Times[Var 0, Var 0], 
	                 Times[Atom "[", Var 0, Atom "]"], 
		         Times[Atom "[", Atom "]"]
		        ])]
in 
    printEqns eqs;
    printEqns (leftcontexts eqs (Var 0));
    printEqns (RegEqns.solveRightlinearEqns (leftcontexts eqs (Var 0)))
end

 Equation system: 
   x0 = (x0;x0 + [;x0;] + [;])

 Equation system: 
   x1 = (1 + [;x1 + x1 + x0;x1)

 Equation system: 
   x1 = (x0 + [)*  

The reduced left context of x0 in L(x0) is ([ + 1 + x0)* = ([ + x0)*



Example 2: (Arbib, Moll, Kfoury p. 137)

local 
    structure Names =
	struct
	    type alph = string
            fun atomName (a : alph) = a : string
            fun varName 0 = "S" | varName 1 = "R" | varName 2 = "E"
	      | varName i = "x"^ Int.toString i 
        end
in 
    structure Grm = ppREGEXP (Names);
end


val S = Var 0
val R = Var 1
val E = Var 2

val eqs = [(0,Plus [R, Times [Atom "<", S, Atom ">"]]),
	  (1,Times [E, Atom "=", E]), 
	  (2,Plus [Times [Atom "<", E, Atom "+", E, Atom ">"], 
		   Atom "a", 
		   Atom "b"])]
Grm.printEqns eqs;
Grm.printEqns (Grammars.leftcontexts eqs (Var 0));
Grm.printEqns (RegEqns.solveRightlinearEqns (Grammars.leftcontexts eqs (Var 0)));

 Equation system: 
   S = (R + <;S;>)
   R = E;=;E
   E = (<;E;+;E;> + a + b)
 val it = () : unit

 Equation system: 
   x3 = (1 + x4 + <;x3)
   x4 = (x5 + E;=;x5)
   x5 = (<;x5 + <;E;+;x5)
 val it = () : unit

 Equation system: 
   x3 = (<)*
   x4 = 0
   x5 = 0

Grm.printEqns (Grammars.leftcontexts eqs (Var 1));
Grm.printEqns (RegEqns.solveRightlinearEqns (Grammars.leftcontexts eqs (Var 1)));

 Equation system: 
   x3 = (x4 + <;x3)
   x4 = (1 + x5 + E;=;x5)
   x5 = (<;x5 + <;E;+;x5)
 val it = () : unit

 Equation system: 
   x3 = (<)*
   x4 = 1
   x5 = 0

Grm.printEqns (Grammars.leftcontexts eqs (Var 2));
Grm.printEqns (RegEqns.solveRightlinearEqns (Grammars.leftcontexts eqs (Var 2)));

 Equation system: 
   x3 = (x4 + <;x3)
   x4 = (x5 + E;=;x5)
   x5 = (1 + <;x5 + <;E;+;x5)
 val it = () : unit

 Equation system: 
   x3 = (<)*;(E;= + 1);(<;E;+ + <)*
   x4 = (E;= + 1);(<;E;+ + <)*
   x5 = (<;E;+ + <)*


Example 3: (Beispiel 3.3.4, Analysealgorithmen S.54/55: (Korrekt))

structure Names3 =
    struct
	type alph = string
        fun atomName (a : alph) = a : string
        fun varName 0 = "S" | varName 1 = "E" 
	  | varName i = "x"^ Int.toString i 
    end;
structure Grm3 = ppREGEXP (Names3);

let
   open Grm3
   open Grammars
   val eqs = [(0,Plus[Times[Var 1,Atom "=",Var 1],Times[Atom "[",Var 0,Atom "]"]]),
              (1,Plus[Times[Atom "[",Var 1,Atom "*",Var 1,Atom "]"],Atom "a",Atom "b"])]
in
    printEqns eqs;
    printEqns (Grammars.leftcontexts eqs (Var 0));
    printEqns (RegEqns.solveRightlinearEqns (leftcontexts eqs (Var 0)));
    printEqns (Grammars.leftcontexts eqs (Var 1));
    printEqns (RegEqns.solveRightlinearEqns (Grammars.leftcontexts eqs (Var 1)))
end

 Equation system: 
   S = (E;=;E + [;S;])
   E = ([;E;*;E;] + a + b)

 Equation system: 
   x2 = (1 + [;x2 + x3 + E;=;x3)
   x3 = ([;x3 + [;E;*;x3)

 Equation system: 
   x2 = ([)*
   x3 = 0

 Equation system: 
   x2 = ([;x2 + x3 + E;=;x3)
   x3 = (1 + [;x3 + [;E;*;x3)

 Equation system: 
   x2 = ([)*;(E;= + 1);([;E;* + [)*
   x3 = ([;E;* + [)*
*)

(* Precomputations of LL(k)-grammars *)


(* Example 2.2.4, Analysealgorithmen *)


local 
    structure Names =
	struct
	    type alph = string
            fun atomName (a : alph) = a : string
            fun varName 0 = "A" | varName 1 = "B" 
	      | varName i = "x"^ Int.toString i 
        end
in 
    structure Grm = ppREGEXP (Names);
end


local
     datatype re = datatype Grm.re
     type alph = Grm.alph
     val eqs = [(0,Times[Star(Times[Atom "a",Var 0]),Var 1]),
                (1,Plus[Times[Atom "b",Var 0],Star(Times[Atom "c",Var 1,Atom "b"])])];
     fun envEmpty i = Set.EmptySet : alph Grammars.t
     fun firstN n = Grammars.first n envEmpty eqs  (* 3 statt 2 dauert schon ewig *)
in 
   fun follsN k s r = Grammars.follows k envEmpty eqs s r
   and flsN k s r = Set.list (follsN k s r)
   and fstN k r =  firstN k r
   and fst k r = Grammars.evalfst k (Grammars.Fst k envEmpty eqs) r
end

(* Vorher (mit (initial k (catenate Ls)) terminierte er nicht.

   flsN 3 (Grm.Times[Grm.Atom "a",Grm.Atom "b",Grm.Atom "c",Grm.Atom "d"]) (Grm.Atom "c"); 

   flsN 1 (Grm.Times[Grm.Atom "a",Grm.Atom "b",Grm.Atom "c",Grm.Atom "d"]) 
          (Grm.nf (Grm.Times[Grm.Atom "b"])); 

   flsN 2 (Grm.Times[Grm.Atom "a",Grm.Star(Grm.Atom "a")]) (Grm.Atom "a"); 
                                   (*  val it = [["a","a"],["a"]] fehlt [] *)
   (* jetzt schleife ?? *)

   flsN 2 (Grm.Times[Grm.Atom "b",Grm.Star(Grm.Atom "a")]) (Grm.Atom "a"); 
           (* falsch [] *)
   flsN 2 (Grm.Atom "a") (Grm.Atom "a");  (* ok, [[]] *)
   flsN 2 (Grm.Times[Grm.Atom "a"]) (Grm.Atom "a");  (* falsch , [] *)
   flsN 2 (Grm.Plus[Grm.Times[Grm.Atom "a",Grm.Atom "b"],Grm.Times[Grm.Atom "b",Grm.Atom "a"]]) 
          (Grm.Atom "a")  (* ok, [["b"]] *)

Test am Beispiel 2.2.4 der Vorlesung: 
A = (aA)*;B
B = (bA + (cBb)*)

   flsN 1 (Grm.Var 0) (Grm.Atom "a");  (* [["a"],["b"],["c"],[]] ok *)
   flsN 1 (Grm.Var 1) (Grm.Atom "a");  (* [["a"],["b"],["c"],[]] ok *)


fun showCat k Ls = Set.list (Cat k (map Set.mk Ls));

showCat 2 [[[1,2,3],[2]],[[]]];
flsN 2 (RegExp.Powers (Grm.Atom "a") 3) (Grm.Atom "a")
flsN 3 (Grm.Star(Grm.Atom "a")) (Grm.Plus[Grm.Atom "a",Grm.Atom "b"])

fun EmptyLenv i = Set.mk[] 

FEHLERSUCHE: Lag an Cat k [] = []: statt [[]]. Aber jetzt ist es wieder zu langsam!
   flsN 1 (Grm.Atom "a")            (Grm.Atom "a");  (* ok, [[]] *)
   flsN 1 (Grm.Times[Grm.Atom "a"]) (Grm.Atom "a");  (* [[]] ok *)
   flsN 1 (Grm.Times[Grm.Atom "a",Grm.Atom "b"]) (Grm.Atom "a");  (* ok [["b"]] *)
   flsN 1 (Grm.Times[Grm.Atom "a",Grm.Unit]) (Grm.Atom "a");  (* [[]] ok *)
   flsN 1 (Grm.Times[Grm.Unit,Grm.Atom "a"]) (Grm.Atom "a");  (* [[]] ok *)

   flsN 1 (RegExp.Powers (Grm.Atom "a") 0) (Grm.Atom "a") (* [] ok *)
   flsN 1 (RegExp.Powers (Grm.Atom "a") 1) (Grm.Atom "a") (* [[]] ok *)
   flsN 1 (RegExp.Powers (Grm.Atom "a") 2) (Grm.Atom "a") (* [[],["a"]] ok *)
   flsN 1 (RegExp.Powers (Grm.Atom "a") 3) (Grm.Atom "a") (* [[],["a"]] ok *)

   flsN 1 (Grm.Star(Grm.Atom "a")) (Grm.Atom "a")  (* [[],["a"]] ok *)
   flsN 1 (Grm.Times[Grm.Star(Grm.Atom "a"),Grm.Atom "b"]) (Grm.Atom "a");  (* [["a",["b"]] ok *)
   flsN 1 (Grm.Times[Grm.Star(Grm.Atom "a"),Grm.Atom "b"]) (Grm.Star(Grm.Atom "a"));  (* ok *)

Set.list (Grammars.evalfst 2 EmptyLenv (Grm.Times[Grm.Atom "a",Grm.Atom "b", Grm.Atom "c"]))
(* ok *)

   flsN 2 (Grm.Atom "a")            (Grm.Atom "a");  (* divergiert??; war mal: [[]] *)

   flsN 1 (Grm.Times[Grm.Atom "a"]) (Grm.Atom "a");  (* [[]] ok *)
   flsN 1 (Grm.Times[Grm.Atom "a",Grm.Atom "b"]) (Grm.Atom "a");  (* ok [["b"]] *)
   flsN 1 (Grm.Times[Grm.Atom "a",Grm.Unit]) (Grm.Atom "a");  (* [[]] ok *)
   flsN 1 (Grm.Times[Grm.Unit,Grm.Atom "a"]) (Grm.Atom "a");  (* [[]] ok *)

Example 2.2.10

local 
    structure Names =
	struct
	    type alph = string
            fun atomName (a : alph) = a : string
            fun varName 0 = "A" | varName 1 = "B" 
	      | varName i = "x"^ Int.toString i 
        end
in 
    structure Grm = ppREGEXP (Names);
end

local
     datatype re = datatype Grm.re
     type alph = Grm.alph
     val (E,D,F) = (Var 0,Var 1,Var 2)
     val (a,e,p,m,l,r) = (Atom "a",Grm.Unit,Atom "+",Atom "x",Atom "<",Atom ">")
     val eqs = [(0,Times[D,Plus[Times[p,D,E],e]]),
                (1,Times[F,Plus[Times[m,F,D],e]]),
                (2,Plus[Times[l,E,r],a])]
     fun envEmpty i = Set.EmptySet : alph Grammars.t
     fun firstN n = Grammars.first n envEmpty eqs  (* 3 statt 2 dauert schon ewig *)
in 
   fun follsN k s r = Grammars.follows k envEmpty eqs s r
   and flsN k s r = Set.list (follsN k s r)
   and fstN k r =  Set.list (firstN k r)
   and fst k r = Set.list (Grammars.evalfst k (Grammars.Fst k envEmpty eqs) r)
end

fst 1 (Grm.Plus[Grm.Times[Grm.Atom "<",Grm.Var 0, Grm.Atom ">"],Grm.Atom "a"])
(* [Atom "<", Atom "a"] ok *)
fst 2 (Grm.Plus[Grm.Times[Grm.Atom "<",Grm.Var 0, Grm.Atom ">"],Grm.Atom "a"])
(* [["<","<"],["<","a"],["a"]] : alph list list *)
fst 3 (Grm.Plus[Grm.Times[Grm.Atom "<",Grm.Var 0, Grm.Atom ">"],Grm.Atom "a"])

flsN 1 (Grm.Var 0) (Grm.Plus[Grm.Times[Grm.Atom "+",Grm.Var 1, Grm.Var 0],Grm.Unit])
(* ok [[],Atom ">"] *)
flsN 1 (Grm.Var 1) (Grm.Plus[Grm.Times[Grm.Atom "+",Grm.Var 1, Grm.Var 0],Grm.Unit])
(* [[">"]] ok *)
flsN 1 (Grm.Var 2) (Grm.Plus[Grm.Times[Grm.Atom "+",Grm.Var 1, Grm.Var 0],Grm.Unit])
(* [[">"]] ok *)

flsN 1 (Grm.Var 0) (Grm.Plus[Grm.Times[Grm.Atom "<",Grm.Var 0, Grm.Atom ">"],Grm.Atom "a"])
(* val it = [[">"],["x"],["<"],["a"],["+"],[]] *)











