(* Kleenean Algebras and Idempotent Semirings. H.Leiss, 20.1.99 *)

signature KA =
    sig 
	type t
	val zero : t
	val sum : t * t -> t
	val unit : t
	val prod : t * t -> t
	val star : t -> t
    end


(* Example: all-pairs shortest paths in a graph *)

structure Path : KA =
    struct
	type t = int
	val SOME zero = Int.maxInt
	val sum = Int.min
	val unit = 0
	fun prod (n,m) = if m=zero orelse n=zero then zero else n+m
	fun star x = 0
    end

(* Example: Conway's non-continous Kleenean algebra, cf. Pratt: Action Locic, JELIA'91 *)

structure Conway : KA =
    struct
	type t = int  (* {0,1,2,3} gives a finite (non-continuous) Kleenean Algebra *)
	val zero = 0
	val unit = 1
	val sum = Int.max
	fun prod (x,y) = if x=0 orelse y=0 then 0 else Int.max (x,y)
	fun star 0 = 1
	  | star 1 = 1
	  | star 2 = 3
	  | star 3 = 3
	  | star _ = raise Size
    end


