signature PPFM = 
  sig
    include FM

    eqtype label 

    val dot : label machine -> string -> unit
    (* write a description of m to the file s in the format of graphviz/dot *)

    val viewgraph : string -> OS.Process.status
    (* display the graph in file s by the graph-drawing function dot. 
       The filename is relative to the directory where SML was called. *)

    val display : label machine -> string -> OS.Process.status
    (* display a picture of the machine m using temporary .dot- and .ps-files with
       basename s. Leaves these files for possible further use. *)

    val displayMinimize : label machine -> string -> OS.Process.status
    (* display a picture of the minimal deterministic machine equivalent to m,
       using temporary .dot- and .ps-files with basename s. 
       Leaves these files for possible further use. *)

    val displayDetermine : label machine -> string -> OS.Process.status
    (* display a picture of the deterministic machine equivalent to m,
       using temporary .dot- and .ps-files with
       basename s. Leaves these files for possible further use. *)

    val displayDetermineMinimize : label machine -> string -> OS.Process.status
        (* a) displays the original automaton m,
           b) displays the deterministic version m2 of m obtained by the powerset construction,
              using sets of states of m to represent states of m2 
           c) displays the deterministic automation m2 with states renamed by numbers
           d) displays the minimal automaton m3 with states as sets of states of m2
           e) displays the minimal automaton equivalent to m with states renamed by numbers.
         *)
  end

