(* File:    ppregeqnsfm.sml 
   Content: Translations from regular recursion equations to finite automata and conversely,
            including the translation between regular expressions and automata.
   Date:    2.3.2000
   Author:  H.Leiss

   5.1.2010 Replaced Compiler.Control.PrintDepth by Control.Print.printDepth
*)

signature PPREGEQNSFM =
    sig
	structure E : PPREGEQNS
	structure M : PPFM sharing type M.label = E.alph 

	type ''a machine = ''a M.machine
        type label = M.label

        datatype re = datatype E.re
	type 'a eqns = 'a re E.eqns
        val printExp : label re -> unit
        val printEqns: label eqns -> unit

	val accept : ''a machine -> ''a list -> bool
        val display : label machine -> string -> OS.Process.status
        
	val retofm : ''a re -> ''a machine
	val retominfm : ''a re -> ''a machine
        val fmtore : ''a machine -> ''a re
        val fmToEqns : ''a machine -> ''a eqns
    end 

(* provide printable names for constants and variables *)

functor ppRegEqnsFm (Names : NAMES) : PPREGEQNSFM = 
    struct
        structure E = ppRegEqns(Names) : PPREGEQNS
        datatype re = datatype E.re
	type 'a eqns = 'a re E.eqns
        val printExp = E.printExp
        and printEqns = E.printEqns

        structure Alphabet =
            struct
                type label = Names.alph
                val toString = Names.atomName 
            end
        structure M = ppFm(Alphabet) : PPFM
	val accept = M.accept
        val display = M.display

        structure EM = RegEqnsFm(structure E = E structure M = M) 

	val retofm = EM.retofm
	val retominfm = EM.retominfm
        val fmtore = EM.fmtore
        val fmToEqns = EM.fmToEqns

        type label = Alphabet.label
	type ''a machine = ''a M.machine

        exception Undefined 
    end

(* Example:

structure StringNames = Names(StringLabels);
structure ppEM = ppRegEqnsFm(StringNames)

open ppEM;

val ndM = M.make{trans = [(0,"a",1),(0,"a",2),
                          (1,"a",0),(1,"b",2),
                          (2,"b",0),(2,"b",1)], 
                 start=[0,2],
                 final=[1,2]};

Control.Print.printDepth := 10; 

val ndE = fmToEqns ndM;
printEqns ndE;

 Equation system: 
   x_0 = (a;x_1 + a;x_2)
   x_1 = (1 + a;x_0 + b;x_2)
   x_2 = (1 + b;x_0 + b;x_1)
   x_3 = (x_0 + x_2)

display ndM "test"

val (3,e)::_ = List.rev(E.solveRightlinearEqns ndE);
val minM = retominfm e
display minM "test";

display (M.minimize(M.determine ndM)) "test2"
*)

(* 
structure RA = ppRegExpFm(StringLabels)
open RA;

val ndM = M.make{trans = [(0,"a",1),(0,"a",2),
                          (1,"a",0),(1,"b",2),
                          (2,"b",0),(2,"b",1)], 
                 start=[0],
                 final=[1,2]};display ndM "test"; (* 3 states *)

val dM = M.determine ndM;display dM "test";       (* 4 states *)
M.show dM;
val it =
  {final=[2,3],
   trans=[(0,"b",1),(0,"a",2),
          (1,"b",1),(1,"a",1),
          (2,"b",3),(2,"a",0),
          (3,"b",3),(3,"a",3)],
   start=[0]}

display (retofm(fmtore ndM)) "test"   (* 91 states! *)
display ((M.minimize o M.determine) (retofm(fmtore dM))) "test"; (* 4 states *)
display (retofm(fmtore dM)) "test";   (* 21 states *)

Control.Print.printDepth := 10; 

fmtore dM;   
- val it =
  Plus
    [Times [Star (Times [Atom "a",Atom "a"]),Atom "a"],
     Times
       [Star (Times [Atom "a",Atom "a"]),Atom "a",Atom "b",
        Star (Plus [Atom "a",Atom "b"])]] : string re
Frueher mit regSys war es nicht ausmultiplziert:
val it =
  Times
    [Atom "a",Star (Times [Atom "a",Atom "a"]),
     Plus [Times [],Times [Atom "b",Star (Plus [Atom "b",Atom "a"])]]]
(* = a(aa)*+a(aa)*b(b+a)* ok *)


Example 5 (Translation from regular expressions (a+b)^*a(a+b)^n to finite machines)

structure T = ppRegExpFm(StringLabels);

val a = T.E.Atom "a" and b = T.E.Atom "b"

fun exp n =  T.E.Times[T.E.Star(T.E.Plus[a,b]),a,T.E.Power (T.E.Plus[a,b]) n]
and mach n = T.M.minimize (T.M.determine (T.retofm (exp n)))
and disp n = T.display (mach n) "test"

T.M.states (mach 2); (*  8 *)
T.M.states (mach 3); (* 16 *)
T.M.states (mach 4); (* 32 *)
T.M.states (mach 5); (* 64 *)
T.M.states (mach 6); (* 128 *)

T.fmtore (mach 0); 
T.fmtore (mach 1); 
T.fmtore (mach 2); (* langsam! *)

disp 3;
disp 4;
*)

