(* Translation of 'a *'b FM's to ('a -> 'b)-Transducers      H.Leiss 1.12.95 
   Edited: 26.02.00
   - renamed TD to TRANSDUCER
   - made functor a structure
   - extracted printable labels and display to pptransducer.sml
*) 

signature TRANSDUCER =
    sig
	type ('a,'b) transducer = ('a * 'b) Fm.machine

	val translate : (''a, ''b) transducer -> ''a list -> ''b list list  
	val accept : (''a, ''b) transducer -> (''a * ''b) list -> bool

	val make : {trans : (int * (''a * ''b) * int) list, 
		    start : int list, 
		    final : int list} -> (''a, ''b) transducer
	val show : (''a, ''b) transducer -> {start : int list, 
					   trans : (int * (''a * ''b) * int) list, 
					   final : int list}

	val Atom  : (''a * ''b) -> (''a, ''b) transducer
	val Plus  : (''a, ''b) transducer list -> (''a, ''b) transducer
	val Times : (''a, ''b) transducer list -> (''a, ''b) transducer
	val Star  : (''a, ''b) transducer -> (''a, ''b) transducer
  end

structure Transducer :> TRANSDUCER =
    struct
	type ('a,'b) transducer = ('a * 'b) Fm.machine

	local
	    fun member n [] = false
	      | member n (x::l) = if n=x then true else member n l
	    and union [] U = U
	      | union (Q::Qs) U = union Qs (disjointAdd Q U)
	    and disjointAdd [] U = U
	      | disjointAdd (q::Q) U = 
		if member q Q then disjointAdd Q U
		else if member q U then disjointAdd Q U 
		     else disjointAdd Q (q::U)
	    and intersects [] Q = false
	      | intersects (p::P) Q = if member p Q then true 
				      else intersects P Q 
					 
	in 
	    fun step_td M (p,W) a =  (* Basisoutput: nur Buchstaben *)
		let 
                    val {trans = transM,...} = Fm.show M
		    fun step_TD (p,W) a ((r,(b,v),q)::Trans) New = 
			if (p,a) = (r,b) 
			    then  step_TD (p,W) a Trans ((q,W @ [v])::New)
			else step_TD (p,W) a Trans New
		      | step_TD (p,W) a [] New = New
		in 
		    step_TD (p,W) a transM [] 
		end

	    and reach M Q (a::A) =
		let 
		    val Qs = map (fn (p,W) => step_td M (p,W) a) Q
		in reach M (union Qs []) A end
	      | reach M Q [] = Q


	    and translate M Inputs = 
		let 
                    val {start = startM, final = finalM, ...} = Fm.show M

		    val Ss = map (fn p => (p,[])) startM
		    val F = finalM
		    val Ws = reach M Ss Inputs
		    fun	select ((q,V)::Vs) Selected =
			if (member q F) then select Vs (V :: Selected) 
			else select Vs Selected
		      | select [] Selected = Selected
		in
		    select Ws []
		end

	    val make = Fm.make : {final:int list, start:int list, 
                                  trans:(int * (''a * ''b) * int) list}
                                 -> (''a,''b) transducer
	    val show = Fm.show 
	    val accept = Fm.accept
		
	    val Atom = Fm.Atom : (''a * ''b) -> (''a,''b) transducer
	    val Plus = Fm.Plus : (''a,''b) transducer list -> (''a,''b) transducer
	    val Times = Fm.Times : (''a,''b) transducer list -> (''a,''b) transducer
	    val Star = Fm.Star : (''a,''b) transducer -> (''a,''b) transducer
	end
    end


(*   **********************   Examples    ******************************

val T = IntBoolTd.make{start = [0], final = [2],
                       trans = [(0,(0,false),0),(0,(1,true),1),(1,(1,false),1),
                         (1,(0,true),1),(1,(0,false),2)]}

TdIB.translate T [0,1,1,0,0,0]

---------------------------------------------        

val T = Td.make{start = [0], final = [2],
                trans = [(0,(0,false),0),(0,(1,true),1),(1,(1,false),1),
                         (1,(0,true),1),(1,(0,false),2)]}

Fa.display T "file";
Td.translate T [0,1,1,0,0,0]
--------------------------------------------
val M2 = Star (Atom (3,"c"));
- show M2;
val it = {final=[0,2],trans=[(1,(3,"c"),2),(1,(3,"c"),1)],start=[0,1]}
  : {final:int list, trans:(int * (int * string) * int) list, start:int list}


val M3 = Star(Plus[Atom (9,"a"),Atom (9,"b")]);
- show M3;
val it =
  {final=[0,2,4],
   trans=[(1,(9,"a"),2),(3,(9,"b"),4),(3,(9,"b"),1),(3,(9,"b"),3),
          (1,(9,"a"),1),(1,(9,"a"),3)],start=[0,1,3]}
  : {final:int list, trans:(int * (int * string) * int) list, start:int list}

- translate M3 [9,9];
val it = [["b","a"],["b","b"],["a","a"],["a","b"]] : string list list


- val M4 = Times [Atom (3,"a"), Atom (4,"b")];
- show M4;
val it =
  {final=[3],trans=[(0,(3,"a"),1),(0,(3,"a"),2),(2,(4,"b"),3)],start=[0]}
  : {final:int list, trans:(int * (int * string) * int) list, start:int list}
- translate M4 [3,4];
val it = [["a","b"]] : string list list


- val M5 = Star M4;
- show M5;
val it =
  {final=[0,4],
   trans=[(0,(3,"a"),3),(0,(3,"a"),2),(1,(3,"a"),2),(1,(3,"a"),3),
          (3,(4,"b"),4),(3,(4,"b"),1)],start=[0,1]}
  : {final:int list, trans:(int * (int * string) * int) list, start:int list}
- translate M5 [3,4,3,4];
val it = [["a","b","a","b"]] : string list list

************************************************************************    *)
